import json
from dataclasses import asdict
from unittest.mock import MagicMock, patch

import requests
import requests_mock
from redis import Redis
from requests.exceptions import HTTPError

from allianceauth import __title__ as AUTH_TITLE
from allianceauth import __url__, __version__
from allianceauth.utils.testing import NoSocketsTestCase

from ...utils import set_logger_to_file
from ..client import (
    DEFAULT_BACKOFF_DELAY,
    DURATION_CONTINGENCY,
    DiscordClient,
    RolesSet,
)
from ..exceptions import DiscordRateLimitExhausted, DiscordTooManyRequestsError
from .factories import (
    TEST_BOT_TOKEN,
    TEST_GUILD_ID,
    TEST_ROLE_ID,
    TEST_USER_ID,
    TEST_USER_NAME,
    create_discord_error_response_unknown_member,
    create_discord_guild_member_object,
    create_discord_guild_object,
    create_discord_role_object,
    create_guild,
    create_guild_member,
    create_matched_role,
    create_role,
    create_user,
)

logger = set_logger_to_file(
    'allianceauth.services.modules.discord.discord_client.client', __file__
)

MODULE_PATH = 'allianceauth.services.modules.discord.discord_client.client'
API_BASE_URL = 'https://discord.com/api/'

TEST_RETRY_AFTER = 3000

DEFAULT_REQUEST_HEADERS = {
    'User-Agent': f'{AUTH_TITLE} ({__url__}, {__version__})',
    'accept': 'application/json',
    'authorization': 'Bot ' + TEST_BOT_TOKEN
}

# default mock redis client to use
mock_redis = MagicMock(**{
    'get.return_value': None,
    'pttl.return_value': -1,
})


# default mock function to simulate sleep
def my_sleep(value):
    if value < 0:
        raise ValueError('sleep length must be non-negative')


class DiscordClientStub(DiscordClient):
    """DiscordClient with stubbed lua wrappers for easier testing."""

    def _redis_set_if_longer(self, name: str, value: str, px: int):
        return True

    def _redis_decr_or_set(self, name: str, value: str, px: int):
        return 5


class TestBasicsAndHelpers(NoSocketsTestCase):

    def test_can_create_object(self):
        # when
        client = DiscordClient(TEST_BOT_TOKEN, mock_redis)
        # then
        self.assertIsInstance(client, DiscordClient)
        self.assertEqual(client.access_token, TEST_BOT_TOKEN)

    def test_should_raise_error_when_trying_to_create_object_without_token(self):
        # when/then
        with self.assertRaises(Exception):
            DiscordClient("", mock_redis)

    def test_repr(self):
        client = DiscordClient(TEST_BOT_TOKEN, mock_redis)
        expected = 'DiscordClient(access_token=...UVWXY)'
        self.assertEqual(repr(client), expected)

    def test_can_set_rate_limiting(self):
        client = DiscordClient(TEST_BOT_TOKEN, mock_redis, is_rate_limited=False)
        self.assertFalse(client.is_rate_limited)

        client = DiscordClient(TEST_BOT_TOKEN, mock_redis, is_rate_limited=True)
        self.assertTrue(client.is_rate_limited)

    @patch(MODULE_PATH + '.get_redis_client', spec=True)
    def test_use_default_redis_if_none_provided(self, mock_get_redis_client):
        # given
        my_redis = MagicMock(spec=Redis)
        mock_get_redis_client.return_value = my_redis
        # when
        client = DiscordClient(TEST_BOT_TOKEN)
        # then
        self.assertEqual(client._redis, my_redis)

    @patch(MODULE_PATH + '.get_redis_client')
    def test_raise_exception_if_redis_client_not_found(self, mock_get_redis_client):
        # given
        mock_get_redis_client.return_value = None
        # when
        with self.assertRaises(RuntimeError):
            DiscordClient(TEST_BOT_TOKEN)


@requests_mock.Mocker()
class TestOtherMethods(NoSocketsTestCase):

    def setUp(self):
        self.client = DiscordClientStub(TEST_BOT_TOKEN, mock_redis)
        self.headers = DEFAULT_REQUEST_HEADERS

    def test_user_get_current(self, requests_mocker):
        user = create_user()
        headers = {
            'accept': 'application/json',
            'authorization': 'Bearer accesstoken'
        }
        requests_mocker.register_uri(
            'GET',
            f'{API_BASE_URL}users/@me',
            request_headers=headers,
            json=asdict(user)
        )
        client = DiscordClientStub('accesstoken', mock_redis)
        result = client.current_user()
        self.assertEqual(result, user)

    def test_get_infos(self, requests_mocker):
        requests_mocker.get(
            f'{API_BASE_URL}guilds/{TEST_GUILD_ID}',
            request_headers=self.headers,
            json=create_discord_guild_object(id=1, name="dummy")
        )
        result = self.client.guild_infos(TEST_GUILD_ID)
        self.assertEqual(result, create_guild(id=1, name="dummy"))


@requests_mock.Mocker()
class TestGuildRoles(NoSocketsTestCase):

    def setUp(self):
        self.url = f'{API_BASE_URL}guilds/{TEST_GUILD_ID}/roles'

    def test_without_cache(self, requests_mocker):
        my_mock_redis = MagicMock(**{
            'get.return_value': None,
            'pttl.return_value': -1,
        })
        requests_mocker.get(
            url=self.url,
            request_headers=DEFAULT_REQUEST_HEADERS,
            json=[create_discord_role_object(1, "alpha")]
        )
        client = DiscordClientStub(TEST_BOT_TOKEN, my_mock_redis)
        result = client.guild_roles(TEST_GUILD_ID, use_cache=False)
        self.assertSetEqual(result, {create_role(id=1, name="alpha")})
        self.assertTrue(my_mock_redis.set.called)

    def test_return_from_cache_if_in_cache(self, requests_mocker):
        my_mock_redis = MagicMock(**{
            'get.return_value': (
                json.dumps([create_discord_role_object(1, "alpha")]).encode('utf8')
            )
        })
        client = DiscordClientStub(TEST_BOT_TOKEN, my_mock_redis)
        result = client.guild_roles(TEST_GUILD_ID)
        self.assertSetEqual(result, {create_role(id=1, name="alpha")})
        self.assertFalse(my_mock_redis.set.called)

    def test_return_from_api_and_save_to_cache_if_not_in_cache(
        self, requests_mocker
    ):
        my_mock_redis = MagicMock(**{
            'get.return_value': None,
            'pttl.return_value': -1,
        })
        requests_mocker.get(
            url=self.url,
            request_headers=DEFAULT_REQUEST_HEADERS,
            json=[create_discord_role_object(1, "alpha")]
        )
        client = DiscordClientStub(TEST_BOT_TOKEN, my_mock_redis)
        result = client.guild_roles(TEST_GUILD_ID)
        self.assertEqual(result, {create_role(id=1, name="alpha")})
        self.assertTrue(my_mock_redis.set.called)

    def test_dont_save_in_cache_if_api_returns_invalid_response_1(
        self, requests_mocker
    ):
        my_mock_redis = MagicMock(**{
            'get.return_value': None,
            'pttl.return_value': -1,
        })
        requests_mocker.get(
            url=self.url,
            request_headers=DEFAULT_REQUEST_HEADERS,
            json={}
        )
        client = DiscordClientStub(TEST_BOT_TOKEN, my_mock_redis)
        with self.assertRaises(RuntimeError):
            client.guild_roles(TEST_GUILD_ID)

    def test_dont_save_in_cache_if_api_returns_invalid_response_2(
        self, requests_mocker
    ):
        expected = "api returns string"
        my_mock_redis = MagicMock(**{
            'get.return_value': None,
            'pttl.return_value': -1,
        })
        requests_mocker.get(
            url=self.url,
            request_headers=DEFAULT_REQUEST_HEADERS,
            json=expected
        )
        client = DiscordClientStub(TEST_BOT_TOKEN, my_mock_redis)
        with self.assertRaises(RuntimeError):
            client.guild_roles(TEST_GUILD_ID)


@requests_mock.Mocker()
class TestGuildMember(NoSocketsTestCase):

    def setUp(self):
        self.client = DiscordClientStub(TEST_BOT_TOKEN, mock_redis)
        self.headers = DEFAULT_REQUEST_HEADERS

    def test_return_guild_member_when_ok(self, requests_mocker):
        requests_mocker.get(
            f'{API_BASE_URL}guilds/{TEST_GUILD_ID}/members/{TEST_USER_ID}',
            request_headers=self.headers,
            json=create_discord_guild_member_object()
        )
        result = self.client.guild_member(TEST_GUILD_ID, TEST_USER_ID)
        expected = create_guild_member()
        self.assertEqual(result, expected)

    def test_return_none_if_member_not_known(self, requests_mocker):
        requests_mocker.get(
            f'{API_BASE_URL}guilds/{TEST_GUILD_ID}/members/{TEST_USER_ID}',
            request_headers=self.headers,
            status_code=404,
            json=create_discord_error_response_unknown_member()
        )
        result = self.client.guild_member(TEST_GUILD_ID, TEST_USER_ID)
        self.assertIsNone(result)

    def test_raise_exception_on_error(self, requests_mocker):
        requests_mocker.get(
            f'{API_BASE_URL}guilds/{TEST_GUILD_ID}/members/{TEST_USER_ID}',
            request_headers=self.headers,
            status_code=500
        )
        with self.assertRaises(HTTPError):
            self.client.guild_member(TEST_GUILD_ID, TEST_USER_ID)


class TestGuildGetName(NoSocketsTestCase):

    @patch(MODULE_PATH + '.DiscordClient.guild_infos', spec=True)
    def test_returns_from_cache_if_found(self, mock_guild_get_infos):
        guild_name = 'Omega'
        my_mock_redis = MagicMock(**{'get.return_value': guild_name.encode('utf8')})
        mock_guild_get_infos.side_effect = RuntimeError
        client = DiscordClientStub(TEST_BOT_TOKEN, my_mock_redis)
        result = client.guild_name(TEST_GUILD_ID)
        self.assertEqual(result, guild_name)
        self.assertTrue(my_mock_redis.get.called)
        self.assertFalse(my_mock_redis.set.called)

    @patch(MODULE_PATH + '.DiscordClient.guild_infos', spec=True)
    def test_fetches_from_server_if_not_found_in_cache_and_stores_in_cache(
        self, mock_guild_get_infos
    ):
        guild_name = 'Omega'
        my_mock_redis = MagicMock(**{'get.return_value': False})
        mock_guild_get_infos.return_value = create_guild(name=guild_name)
        client = DiscordClientStub(TEST_BOT_TOKEN, my_mock_redis)
        result = client.guild_name(TEST_GUILD_ID)
        self.assertEqual(result, guild_name)
        self.assertTrue(my_mock_redis.get.called)
        self.assertTrue(my_mock_redis.set.called)

    @patch(MODULE_PATH + '.DiscordClient.guild_infos', spec=True)
    def test_fetches_from_server_if_asked_to_ignore_cache_and_stores_in_cache(
        self, mock_guild_get_infos
    ):
        guild_name = 'Omega'
        my_mock_redis = MagicMock(**{'get.return_value': False})
        mock_guild_get_infos.return_value = create_guild(name=guild_name)
        client = DiscordClientStub(TEST_BOT_TOKEN, my_mock_redis)
        result = client.guild_name(TEST_GUILD_ID, use_cache=False)
        self.assertFalse(my_mock_redis.get.called)
        self.assertEqual(result, guild_name)
        self.assertTrue(my_mock_redis.set.called)

    @patch(MODULE_PATH + '.DiscordClient.guild_infos', spec=True)
    def test_return_empty_if_not_found_in_cache_and_server_has_error(
        self, mock_guild_get_infos
    ):
        # given
        my_mock_redis = MagicMock(**{'get.return_value': False})
        mock_guild_get_infos.side_effect = HTTPError
        client = DiscordClientStub(TEST_BOT_TOKEN, my_mock_redis)
        # when
        result = client.guild_name(TEST_GUILD_ID)
        # then
        self.assertEqual(result, '')
        self.assertTrue(my_mock_redis.get.called)
        self.assertFalse(my_mock_redis.set.called)


@requests_mock.Mocker()
class TestCreateGuildRole(NoSocketsTestCase):

    def setUp(self):
        self.request_url = f'{API_BASE_URL}guilds/{TEST_GUILD_ID}/roles'
        self.my_mock_redis = MagicMock(**{
            'get.return_value': None,
            'pttl.return_value': -1,
        })
        self.client = DiscordClientStub(TEST_BOT_TOKEN, self.my_mock_redis)

    def test_guild_create_role_normal(self, requests_mocker):
        role_name_input = 'x' * 120
        role_name_used = 'x' * 100
        expected = create_discord_role_object(id=1, name=role_name_used)

        requests_mocker.post(
            self.request_url,
            request_headers=DEFAULT_REQUEST_HEADERS,
            text=json.dumps(expected),
        )
        result = self.client.create_guild_role(
            guild_id=TEST_GUILD_ID, role_name=role_name_input
        )
        self.assertEqual(result, create_role(id=1, name=role_name_used))
        self.assertTrue(self.my_mock_redis.delete.called)

    def test_guild_create_role_empty_response(self, requests_mocker):
        requests_mocker.post(
            self.request_url,
            request_headers=DEFAULT_REQUEST_HEADERS,
            text=json.dumps({}),
        )
        result = self.client.create_guild_role(
            guild_id=TEST_GUILD_ID, role_name='dummy'
        )
        self.assertIsNone(result)
        self.assertFalse(self.my_mock_redis.delete.called)


@requests_mock.Mocker()
class TestGuildDeleteRole(NoSocketsTestCase):

    def setUp(self):
        self.request_url = \
            f'{API_BASE_URL}guilds/{TEST_GUILD_ID}/roles/{TEST_ROLE_ID}'
        self.my_mock_redis = MagicMock(**{
            'get.return_value': None,
            'pttl.return_value': -1,
        })
        self.client = DiscordClientStub(TEST_BOT_TOKEN, self.my_mock_redis)

    def test_guild_delete_role_success(self, requests_mocker):
        requests_mocker.delete(
            self.request_url,
            request_headers=DEFAULT_REQUEST_HEADERS,
            status_code=204
        )
        result = self.client.delete_guild_role(
            guild_id=TEST_GUILD_ID, role_id=TEST_ROLE_ID
        )
        self.assertTrue(result)
        self.assertTrue(self.my_mock_redis.delete.called)

    def test_guild_delete_role_failed(self, requests_mocker):
        requests_mocker.delete(
            self.request_url,
            request_headers=DEFAULT_REQUEST_HEADERS,
            status_code=200
        )
        result = self.client.delete_guild_role(
            guild_id=TEST_GUILD_ID, role_id=TEST_ROLE_ID
        )
        self.assertFalse(result)
        self.assertFalse(self.my_mock_redis.delete.called)


@requests_mock.Mocker()
class TestGuildAddMember(NoSocketsTestCase):

    def setUp(self):
        self.access_token = 'accesstoken'
        self.headers = DEFAULT_REQUEST_HEADERS
        self.request_url = \
            f'{API_BASE_URL}guilds/{TEST_GUILD_ID}/members/{TEST_USER_ID}'
        self.client = DiscordClientStub(TEST_BOT_TOKEN, mock_redis)

    def test_create_new_without_params(self, requests_mocker):

        def data_matcher(request):
            expected = {'access_token': self.access_token}
            return (json.loads(request.text) == expected)

        requests_mocker.register_uri(
            'PUT',
            self.request_url,
            request_headers=self.headers,
            additional_matcher=data_matcher,
            status_code=201,
        )
        result = self.client.add_guild_member(
            guild_id=TEST_GUILD_ID,
            user_id=TEST_USER_ID,
            access_token=self.access_token
        )
        self.assertTrue(result)

    def test_create_existing_without_params(self, requests_mocker):

        def data_matcher(request):
            expected = {'access_token': self.access_token}
            return (json.loads(request.text) == expected)

        requests_mocker.register_uri(
            'PUT',
            self.request_url,
            request_headers=self.headers,
            additional_matcher=data_matcher,
            status_code=204,
        )
        result = self.client.add_guild_member(
            guild_id=TEST_GUILD_ID,
            user_id=TEST_USER_ID,
            access_token=self.access_token
        )
        self.assertIsNone(result)

    def test_create_failed_without_params(self, requests_mocker):

        def data_matcher(request):
            expected = {'access_token': self.access_token}
            return (json.loads(request.text) == expected)

        requests_mocker.register_uri(
            'PUT',
            self.request_url,
            request_headers=self.headers,
            additional_matcher=data_matcher,
            status_code=200,
        )
        result = self.client.add_guild_member(
            guild_id=TEST_GUILD_ID,
            user_id=TEST_USER_ID,
            access_token=self.access_token
        )
        self.assertFalse(result)

    def test_create_new_with_roles(self, requests_mocker):

        role_ids = [1, 2]

        def data_matcher(request):
            expected = {
                'access_token': self.access_token,
                'roles': role_ids
            }
            return (json.loads(request.text) == expected)

        requests_mocker.register_uri(
            'PUT',
            self.request_url,
            request_headers=self.headers,
            additional_matcher=data_matcher,
            status_code=201,
        )
        result = self.client.add_guild_member(
            guild_id=TEST_GUILD_ID,
            user_id=TEST_USER_ID,
            access_token=self.access_token,
            role_ids=role_ids
        )
        self.assertTrue(result)

    def test_raise_exception_on_invalid_roles(self, requests_mocker):
        with self.assertRaises(ValueError):
            self.client.add_guild_member(
                guild_id=TEST_GUILD_ID,
                user_id=TEST_USER_ID,
                access_token=self.access_token,
                role_ids=['abc', 'def']
            )

    def test_create_new_with_nick(self, requests_mocker):

        nick_input = 'x' * 50
        nick_used = 'x' * 32

        def data_matcher(request):
            expected = {
                'access_token': self.access_token,
                'nick': nick_used
            }
            return (json.loads(request.text) == expected)

        requests_mocker.register_uri(
            'PUT',
            self.request_url,
            request_headers=self.headers,
            additional_matcher=data_matcher,
            status_code=201,
        )
        result = self.client.add_guild_member(
            guild_id=TEST_GUILD_ID,
            user_id=TEST_USER_ID,
            access_token=self.access_token,
            nick=nick_input
        )
        self.assertTrue(result)


@requests_mock.Mocker()
class TestGuildModifyMember(NoSocketsTestCase):

    def setUp(self):
        self.access_token = 'accesstoken'
        self.headers = DEFAULT_REQUEST_HEADERS.copy()
        self.headers['content-type'] = 'application/json'
        self.request_url = \
            f'{API_BASE_URL}guilds/{TEST_GUILD_ID}/members/{TEST_USER_ID}'
        self.client = DiscordClientStub(TEST_BOT_TOKEN, mock_redis)

    def test_can_update_roles(self, requests_mocker):
        role_ids = [1, 2]

        def data_matcher(request):
            expected = {'roles': role_ids}
            return (json.loads(request.text) == expected)

        requests_mocker.register_uri(
            'patch',
            self.request_url,
            request_headers=self.headers,
            additional_matcher=data_matcher,
            status_code=204,
        )
        result = self.client.modify_guild_member(
            guild_id=TEST_GUILD_ID,
            user_id=TEST_USER_ID,
            role_ids=role_ids
        )
        self.assertTrue(result)

    def test_can_update_nick(self, requests_mocker):
        nick_input = 'x' * 50
        nick_used = 'x' * 32

        def data_matcher(request):
            expected = {'nick': nick_used}
            return (json.loads(request.text) == expected)

        requests_mocker.register_uri(
            'patch',
            self.request_url,
            request_headers=self.headers,
            additional_matcher=data_matcher,
            status_code=204,
        )
        result = self.client.modify_guild_member(
            guild_id=TEST_GUILD_ID,
            user_id=TEST_USER_ID,
            nick=nick_input
        )
        self.assertTrue(result)

    def test_can_update_roles_and_nick(self, requests_mocker):
        role_ids = [1, 2]

        def data_matcher(request):
            expected = {
                'roles': role_ids,
                'nick': TEST_USER_NAME
            }
            return (json.loads(request.text) == expected)

        requests_mocker.register_uri(
            'patch',
            self.request_url,
            request_headers=self.headers,
            additional_matcher=data_matcher,
            status_code=204,
        )
        result = self.client.modify_guild_member(
            guild_id=TEST_GUILD_ID,
            user_id=TEST_USER_ID,
            role_ids=role_ids,
            nick=TEST_USER_NAME,
        )
        self.assertTrue(result)

    def test_returns_none_if_member_is_unknown(self, requests_mocker):

        def data_matcher(request):
            expected = {'nick': TEST_USER_NAME}
            return (json.loads(request.text) == expected)

        requests_mocker.register_uri(
            'patch',
            self.request_url,
            request_headers=self.headers,
            additional_matcher=data_matcher,
            status_code=404,
            json=create_discord_error_response_unknown_member()
        )
        result = self.client.modify_guild_member(
            guild_id=TEST_GUILD_ID,
            user_id=TEST_USER_ID,
            nick=TEST_USER_NAME
        )
        self.assertIsNone(result)

    def test_returns_false_if_unsuccessful(self, requests_mocker):

        def data_matcher(request):
            expected = {'nick': TEST_USER_NAME}
            return (json.loads(request.text) == expected)

        requests_mocker.register_uri(
            'patch',
            self.request_url,
            request_headers=self.headers,
            additional_matcher=data_matcher,
            status_code=200,
        )
        result = self.client.modify_guild_member(
            guild_id=TEST_GUILD_ID,
            user_id=TEST_USER_ID,
            nick=TEST_USER_NAME
        )
        self.assertFalse(result)

    def test_raise_exception_on_invalid_roles(self, requests_mocker):
        with self.assertRaises(ValueError):
            self.client.modify_guild_member(
                guild_id=TEST_GUILD_ID,
                user_id=TEST_USER_ID,
                role_ids=['abc', 'def']
            )

    def test_raise_exception_if_role_ids_not_list_like(self, requests_mocker):
        with self.assertRaises(TypeError):
            self.client.modify_guild_member(
                guild_id=TEST_GUILD_ID,
                user_id=TEST_USER_ID,
                role_ids='I am not a list'
            )

    def test_raise_exception_on_missing_params(
        self, requests_mocker
    ):
        with self.assertRaises(ValueError):
            self.client.modify_guild_member(
                guild_id=TEST_GUILD_ID,
                user_id=TEST_USER_ID
            )


class TestGuildRemoveMember(NoSocketsTestCase):

    def setUp(self):
        self.headers = DEFAULT_REQUEST_HEADERS
        self.request_url = \
            f'{API_BASE_URL}guilds/{TEST_GUILD_ID}/members/{TEST_USER_ID}'
        self.client = DiscordClientStub(TEST_BOT_TOKEN, mock_redis)

    @requests_mock.Mocker()
    def test_returns_true_on_success(self, requests_mocker):
        requests_mocker.register_uri(
            'DELETE',
            self.request_url,
            request_headers=self.headers,
            status_code=204
        )
        result = self.client.remove_guild_member(
            guild_id=TEST_GUILD_ID, user_id=TEST_USER_ID
        )
        self.assertTrue(result)

    @requests_mock.Mocker()
    def test_returns_none_if_member_unknown(self, requests_mocker):
        requests_mocker.register_uri(
            'DELETE',
            self.request_url,
            request_headers=self.headers,
            status_code=404,
            json=create_discord_error_response_unknown_member()
        )
        result = self.client.remove_guild_member(
            guild_id=TEST_GUILD_ID, user_id=TEST_USER_ID
        )
        self.assertIsNone(result)

    @requests_mock.Mocker()
    def test_raise_exception_on_404_if_member_known(self, requests_mocker):
        requests_mocker.register_uri(
            'DELETE',
            self.request_url,
            request_headers=self.headers,
            status_code=404,
            json={}
        )
        with self.assertRaises(HTTPError):
            self.client.remove_guild_member(
                guild_id=TEST_GUILD_ID, user_id=TEST_USER_ID
            )

    @requests_mock.Mocker()
    def test_raise_exception_on_404_if_no_api_response(self, requests_mocker):
        requests_mocker.register_uri(
            'DELETE',
            self.request_url,
            request_headers=self.headers,
            status_code=404
        )
        with self.assertRaises(HTTPError):
            self.client.remove_guild_member(
                guild_id=TEST_GUILD_ID, user_id=TEST_USER_ID
            )

    @requests_mock.Mocker()
    def test_returns_false_when_not_successful(self, requests_mocker):
        requests_mocker.register_uri(
            'DELETE',
            self.request_url,
            request_headers=self.headers,
            status_code=200
        )
        result = self.client.remove_guild_member(
            guild_id=TEST_GUILD_ID, user_id=TEST_USER_ID
        )
        self.assertFalse(result)


class TestGuildMemberAddRole(NoSocketsTestCase):

    def setUp(self):
        self.headers = DEFAULT_REQUEST_HEADERS
        self.request_url = (
            f'{API_BASE_URL}guilds/{TEST_GUILD_ID}/members/{TEST_USER_ID}'
            f'/roles/{TEST_ROLE_ID}'
        )
        self.client = DiscordClientStub(TEST_BOT_TOKEN, mock_redis)

    @requests_mock.Mocker()
    def test_returns_true_on_success(self, requests_mocker):
        requests_mocker.register_uri(
            'PUT',
            self.request_url,
            request_headers=self.headers,
            status_code=204
        )
        result = self.client.add_guild_member_role(
            guild_id=TEST_GUILD_ID, user_id=TEST_USER_ID, role_id=TEST_ROLE_ID
        )
        self.assertTrue(result)

    @requests_mock.Mocker()
    def test_return_none_if_member_not_known(self, requests_mocker):
        requests_mocker.register_uri(
            'PUT',
            self.request_url,
            request_headers=self.headers,
            status_code=404,
            json=create_discord_error_response_unknown_member()
        )
        result = self.client.add_guild_member_role(
            guild_id=TEST_GUILD_ID, user_id=TEST_USER_ID, role_id=TEST_ROLE_ID
        )
        self.assertIsNone(result)

    @requests_mock.Mocker()
    def test_returns_false_when_not_successful(self, requests_mocker):
        requests_mocker.register_uri(
            'PUT',
            self.request_url,
            request_headers=self.headers,
            status_code=200
        )
        result = self.client.add_guild_member_role(
            guild_id=TEST_GUILD_ID, user_id=TEST_USER_ID, role_id=TEST_ROLE_ID
        )
        self.assertFalse(result)


class TestGuildMemberRemoveRole(NoSocketsTestCase):

    def setUp(self):
        self.headers = DEFAULT_REQUEST_HEADERS
        self.request_url = (
            f'{API_BASE_URL}guilds/{TEST_GUILD_ID}/members/{TEST_USER_ID}'
            f'/roles/{TEST_ROLE_ID}'
        )
        self.client = DiscordClientStub(TEST_BOT_TOKEN, mock_redis)

    @requests_mock.Mocker()
    def test_returns_true_on_success(self, requests_mocker):
        requests_mocker.register_uri(
            'DELETE',
            self.request_url,
            request_headers=self.headers,
            status_code=204
        )
        result = self.client.remove_guild_member_role(
            guild_id=TEST_GUILD_ID, user_id=TEST_USER_ID, role_id=TEST_ROLE_ID
        )
        self.assertTrue(result)

    @requests_mock.Mocker()
    def test_return_none_if_member_not_known(self, requests_mocker):
        requests_mocker.register_uri(
            'DELETE',
            self.request_url,
            request_headers=self.headers,
            status_code=404,
            json=create_discord_error_response_unknown_member()
        )
        result = self.client.remove_guild_member_role(
            guild_id=TEST_GUILD_ID, user_id=TEST_USER_ID, role_id=TEST_ROLE_ID
        )
        self.assertIsNone(result)

    @requests_mock.Mocker()
    def test_returns_false_when_not_successful(self, requests_mocker):
        requests_mocker.register_uri(
            'DELETE',
            self.request_url,
            request_headers=self.headers,
            status_code=200
        )
        result = self.client.remove_guild_member_role(
            guild_id=TEST_GUILD_ID, user_id=TEST_USER_ID, role_id=TEST_ROLE_ID
        )
        self.assertFalse(result)


@patch(MODULE_PATH + '.DiscordClient.guild_roles', spec=True)
@patch(MODULE_PATH + '.DiscordClient.guild_member', spec=True)
class TestGuildMemberRoles(NoSocketsTestCase):
    def test_should_return_member_roles(
        self, mock_guild_member, mock_guild_roles
    ):
        # given
        role_a = create_role(id=1)
        role_b = create_role(id=2)
        mock_guild_member.return_value = create_guild_member(roles=[1])
        mock_guild_roles.return_value = {role_a, role_b}
        client = DiscordClientStub(TEST_BOT_TOKEN, mock_redis)
        # when
        result = client.guild_member_roles(TEST_GUILD_ID, TEST_USER_ID)
        # then
        self.assertSetEqual(result.ids(), {1})

    def test_should_return_none_when_no_member(
        self, mock_guild_member, mock_guild_roles
    ):
        # given
        mock_guild_member.return_value = None
        client = DiscordClientStub(TEST_BOT_TOKEN, mock_redis)
        # when
        result = client.guild_member_roles(TEST_GUILD_ID, TEST_USER_ID)
        # then
        self.assertIsNone(result)

    def test_should_raise_exception_if_member_has_unknown_roles(
        self, mock_guild_member, mock_guild_roles
    ):
        # given
        role_a = create_role(id=1)
        role_b = create_role(id=2)
        mock_guild_member.return_value = create_guild_member(roles=[1, 99])
        mock_guild_roles.return_value = {role_a, role_b}
        client = DiscordClientStub(TEST_BOT_TOKEN, mock_redis)
        # when/then
        roles = client.guild_member_roles(TEST_GUILD_ID, TEST_USER_ID)
        self.assertEqual(roles, RolesSet([role_a]))

    # TODO: Re-enable after adding Discord general error handling
    # def test_should_raise_exception_if_member_info_is_invalid(
    #     self, mock_guild_member, mock_guild_roles
    # ):
    #     # given
    #     member_info = create_guild_member(roles=[1])
    #     del(member_info["roles"])
    #     mock_guild_member.return_value = member_info
    #     mock_guild_roles.return_value = [ROLE_ALPHA, ROLE_BRAVO]
    #     client = DiscordClientStub(TEST_BOT_TOKEN, mock_redis)
    #     # when/then
    #     with self.assertRaises(RuntimeError):
    #         client.guild_member_roles(TEST_GUILD_ID, TEST_USER_ID)

    def test_refresh_guild_roles_user_roles_dont_not_match(
        self, mock_guild_member, mock_guild_roles
    ):
        def my_guild_roles(guild_id, use_cache=True):
            if use_cache:
                return {role_a}
            return {role_a, role_b}

        # given
        role_a = create_role(id=1)
        role_b = create_role(id=2)
        mock_guild_member.return_value = create_guild_member(roles=[1, 2])
        mock_guild_roles.side_effect = my_guild_roles
        client = DiscordClientStub(TEST_BOT_TOKEN, mock_redis)
        # when
        result = client.guild_member_roles(TEST_GUILD_ID, TEST_USER_ID)
        # then
        self.assertSetEqual(result.ids(), {1, 2})
        self.assertEqual(mock_guild_roles.call_count, 2)


class TestMatchGuildRolesToName(NoSocketsTestCase):

    def setUp(self):
        self.url = f'{API_BASE_URL}guilds/{TEST_GUILD_ID}/roles'

    @requests_mock.Mocker()
    def test_return_role_if_known(self, requests_mocker):
        # given
        my_mock_redis = MagicMock(**{
            'get.return_value': None,
            'pttl.return_value': -1,
        })
        requests_mocker.get(
            url=self.url,
            request_headers=DEFAULT_REQUEST_HEADERS,
            json=[create_discord_role_object(1, "alpha")]
        )
        client = DiscordClientStub(TEST_BOT_TOKEN, my_mock_redis)
        # when
        result = client.match_role_from_name(TEST_GUILD_ID, "alpha")
        # then
        self.assertEqual(result, create_role(id=1, name="alpha"))

    @requests_mock.Mocker()
    def test_return_empty_none_if_not_known(self, requests_mocker):
        # given
        my_mock_redis = MagicMock(**{
            'get.return_value': None,
            'pttl.return_value': -1,
        })
        requests_mocker.get(
            url=self.url,
            request_headers=DEFAULT_REQUEST_HEADERS,
            json=[create_discord_role_object(1, "alpha")]
        )
        client = DiscordClientStub(TEST_BOT_TOKEN, my_mock_redis)
        # when
        result = client.match_role_from_name(TEST_GUILD_ID, "unknown")
        # then
        self.assertIsNone(result)


@patch(MODULE_PATH + '.DiscordClient.create_guild_role', spec=True)
@patch(MODULE_PATH + '.DiscordClient.guild_roles', spec=True)
class TestMatchOrCreateGuildRolesToName(NoSocketsTestCase):

    def test_return_role_if_known(
        self, mock_guild_get_roles, mock_guild_create_role,
    ):
        # given
        role_a = create_role(name="alpha")
        role_b = create_role(name="bravo")
        mock_guild_get_roles.return_value = {role_a, role_b}
        client = DiscordClientStub(TEST_BOT_TOKEN, mock_redis)
        # when
        result = client.match_or_create_role_from_name(TEST_GUILD_ID, "alpha")
        # then
        self.assertEqual(result, create_matched_role(role_a, False))
        self.assertFalse(mock_guild_create_role.called)

    def test_create_role_if_not_known_and_return_it(
        self, mock_guild_get_roles, mock_guild_create_role,
    ):
        # given
        role_existing = create_role()
        role_new = create_role(name='echo')
        mock_guild_get_roles.return_value = {role_existing}
        mock_guild_create_role.return_value = role_new
        client = DiscordClientStub(TEST_BOT_TOKEN, mock_redis)
        # when
        result = client.match_or_create_role_from_name(TEST_GUILD_ID, 'echo')
        # then
        expected = (role_new, True)
        self.assertEqual(result, expected)
        self.assertTrue(mock_guild_create_role.called)

    @patch(MODULE_PATH + '.DISCORD_DISABLE_ROLE_CREATION', True)
    def test_return_none_if_role_creation_is_disabled(
        self, mock_guild_get_roles, mock_guild_create_role,
    ):
        # given
        mock_guild_get_roles.return_value = {create_role()}
        client = DiscordClientStub(TEST_BOT_TOKEN, mock_redis)
        # when
        result = client.match_or_create_role_from_name(TEST_GUILD_ID, 'echo')
        # then
        self.assertIsNone(result[0])
        self.assertFalse(result[1])
        self.assertFalse(mock_guild_create_role.called)

    def test_raise_exception_if_name_has_invalid_type(
        self, mock_guild_get_roles, mock_guild_create_role,
    ):
        mock_guild_get_roles.return_value = {}
        client = DiscordClientStub(TEST_BOT_TOKEN, mock_redis)
        with self.assertRaises(TypeError):
            client.match_or_create_role_from_name(TEST_GUILD_ID, 666)


@patch(MODULE_PATH + '.DiscordClient.create_guild_role', spec=True)
@patch(MODULE_PATH + '.DiscordClient.guild_roles', spec=True)
class TestMatchOrCreateGuildRolesToNames(NoSocketsTestCase):

    def test_return_roles_if_known(
        self, mock_guild_get_roles, mock_guild_create_role,
    ):
        # given
        role_a = create_role(name='alpha')
        role_b = create_role(name='bravo')
        mock_guild_get_roles.return_value = {role_a, role_b}
        client = DiscordClientStub(TEST_BOT_TOKEN, mock_redis)
        # when
        result = client.match_or_create_roles_from_names(
            TEST_GUILD_ID, ['alpha', 'bravo']
        )
        # then
        expected = [create_matched_role(role_a), create_matched_role(role_b)]
        self.assertEqual(
            RolesSet.create_from_matched_roles(result),
            RolesSet.create_from_matched_roles(expected)
        )
        self.assertFalse(mock_guild_create_role.called)

    def test_return_roles_if_known_and_create_if_not_known(
        self, mock_guild_get_roles, mock_guild_create_role,
    ):
        role_existing = create_role(name='alpha')
        role_new = create_role(name='echo')
        mock_guild_get_roles.return_value = {role_existing}
        mock_guild_create_role.return_value = role_new
        client = DiscordClientStub(TEST_BOT_TOKEN, mock_redis)
        # when
        result = client.match_or_create_roles_from_names(
            TEST_GUILD_ID, ['alpha', 'echo']
        )
        # then
        expected = [
            create_matched_role(role_existing), create_matched_role(role_new, True)
        ]
        self.assertEqual(
            RolesSet.create_from_matched_roles(result),
            RolesSet.create_from_matched_roles(expected)
        )
        self.assertTrue(mock_guild_create_role.called)

    @patch(MODULE_PATH + '.DISCORD_DISABLE_ROLE_CREATION', True)
    def test_do_not_create_unknown_roles_when_so_configured(
        self, mock_guild_get_roles, mock_guild_create_role,
    ):
        # given
        role_a = create_role(name='alpha')
        mock_guild_get_roles.return_value = {role_a}
        client = DiscordClientStub(TEST_BOT_TOKEN, mock_redis)
        # when
        result = client.match_or_create_roles_from_names(
            TEST_GUILD_ID, ['alpha', 'echo']
        )
        # then
        expected = [create_matched_role(role_a)]
        self.assertEqual(
            RolesSet.create_from_matched_roles(result),
            RolesSet.create_from_matched_roles(expected)
        )
        self.assertFalse(mock_guild_create_role.called)

    def test_consolidate_roles_of_same_name(
        self, mock_guild_get_roles, mock_guild_create_role,
    ):
        # given
        role_a = create_role(name='alpha')
        mock_guild_get_roles.return_value = {role_a}
        client = DiscordClientStub(TEST_BOT_TOKEN, mock_redis)
        # when
        result = client.match_or_create_roles_from_names(
            TEST_GUILD_ID, ['alpha', 'alpha']
        )
        # then
        expected = [create_matched_role(role_a)]
        self.assertEqual(
            RolesSet.create_from_matched_roles(result),
            RolesSet.create_from_matched_roles(expected)
        )
        self.assertFalse(mock_guild_create_role.called)

    def test_consolidate_roles_of_same_name_after_sanitation(
        self, mock_guild_get_roles, mock_guild_create_role,
    ):
        # given
        base_role_name = 'x' * 100
        role = create_role(name=base_role_name)
        role_names = [base_role_name + '1', base_role_name + '2']
        mock_guild_get_roles.return_value = {role}
        mock_guild_create_role.return_value = role
        client = DiscordClientStub(TEST_BOT_TOKEN, mock_redis)
        # when
        result = client.match_or_create_roles_from_names(TEST_GUILD_ID, role_names)
        # then
        expected = [create_matched_role(role)]
        self.assertEqual(
            RolesSet.create_from_matched_roles(result),
            RolesSet.create_from_matched_roles(expected)
        )
        self.assertFalse(mock_guild_create_role.called)

    def test_should_return_empty_list_when_no_names_given(
        self, mock_guild_get_roles, mock_guild_create_role,
    ):
        # given
        mock_guild_get_roles.return_value = {create_role()}
        client = DiscordClientStub(TEST_BOT_TOKEN, mock_redis)
        # when
        result = client.match_or_create_roles_from_names(TEST_GUILD_ID, [])
        # then
        self.assertListEqual(result, [])
        self.assertFalse(mock_guild_create_role.called)


class TestMatchOrCreateRolesFromNames2(NoSocketsTestCase):
    def test_should_return_roles(self):
        # given
        role_a = create_role(name='alpha')
        role_b = create_role(name='bravo')
        client = DiscordClient(TEST_BOT_TOKEN, mock_redis)
        matched_roles = [
            create_matched_role(role_a, False), create_matched_role(role_b, False)
        ]
        client.match_or_create_roles_from_names = (
            lambda *args, **kwargs: matched_roles
        )
        # when
        result = client.match_or_create_roles_from_names_2(
            TEST_GUILD_ID, ['alpha', 'bravo']
        )
        # then
        self.assertEqual(result, RolesSet([role_a, role_b]))

    def test_should_return_empty_roles(self):
        # given
        client = DiscordClient(TEST_BOT_TOKEN, mock_redis)
        client.match_or_create_roles_from_names = (
            lambda *args, **kwargs: []
        )
        # when
        result = client.match_or_create_roles_from_names_2(TEST_GUILD_ID, [])
        # then
        self.assertEqual(result, RolesSet([]))


class TestApiRequestBasics(NoSocketsTestCase):

    def setUp(self):
        self.client = DiscordClient(TEST_BOT_TOKEN, mock_redis)

    @patch(MODULE_PATH + '.requests', spec=requests)
    def test_raises_exception_on_invalid_method(self, mock_requests):
        with self.assertRaises(ValueError):
            self.client._api_request('xxx', 'users/@me')


@patch(MODULE_PATH + '.DiscordClient._redis_decr_or_set', spec=True)
@requests_mock.Mocker()
class TestRateLimitMechanic(NoSocketsTestCase):

    my_role_api = create_discord_role_object(1, "alpha")
    my_role_obj = create_role(id=1, name="alpha")

    @staticmethod
    def my_redis_pttl(name: str):
        if name == DiscordClient._KEY_GLOBAL_BACKOFF_UNTIL:
            return -1
        return TEST_RETRY_AFTER

    def test_proceed_if_requests_remaining(
        self, mock_redis_decr_or_set, requests_mocker
    ):
        requests_mocker.post(
            f'{API_BASE_URL}guilds/{TEST_GUILD_ID}/roles', json=self.my_role_api
        )
        my_mock_redis = MagicMock(**{'pttl.side_effect': self.my_redis_pttl})
        mock_redis_decr_or_set.return_value = 5
        client = DiscordClient(TEST_BOT_TOKEN, my_mock_redis)
        result = client.create_guild_role(
            guild_id=TEST_GUILD_ID, role_name=self.my_role_api['name']
        )
        self.assertEqual(result, self.my_role_obj)

    @patch(MODULE_PATH + '.sleep', spec=True)
    def test_wait_if_reset_happens_soon(
        self, requests_mocker, mock_sleep, mock_redis_decr_or_set
    ):
        counter = 0

        def my_redis_pttl_2(name: str):
            if name == DiscordClient._KEY_GLOBAL_BACKOFF_UNTIL:
                return -1
            else:
                return 100

        def my_redis_decr_or_set(**kwargs):
            nonlocal counter
            counter += 1

            if counter < 2:
                return -1
            else:
                return 5

        requests_mocker.post(
            f'{API_BASE_URL}guilds/{TEST_GUILD_ID}/roles', json=self.my_role_api
        )
        mock_sleep.side_effect = my_sleep
        my_mock_redis = MagicMock(**{'pttl.side_effect': my_redis_pttl_2})
        mock_redis_decr_or_set.side_effect = my_redis_decr_or_set
        client = DiscordClient(TEST_BOT_TOKEN, my_mock_redis)

        result = client.create_guild_role(
            guild_id=TEST_GUILD_ID, role_name=self.my_role_api['name']
        )
        self.assertEqual(result, self.my_role_obj)
        self.assertTrue(mock_sleep.called)

    @patch(MODULE_PATH + '.sleep', spec=True)
    def test_wait_if_reset_happens_soon_and_sleep_must_not_be_negative(
        self, requests_mocker, mock_sleep, mock_redis_decr_or_set
    ):
        counter = 0

        def my_redis_pttl_2(name: str):
            if name == DiscordClient._KEY_GLOBAL_BACKOFF_UNTIL:
                return -1
            else:
                return -1

        def my_redis_decr_or_set(**kwargs):
            nonlocal counter
            counter += 1

            if counter < 2:
                return -1
            else:
                return 5

        requests_mocker.post(
            f'{API_BASE_URL}guilds/{TEST_GUILD_ID}/roles', json=self.my_role_api
        )
        mock_sleep.side_effect = my_sleep
        my_mock_redis = MagicMock(**{'pttl.side_effect': my_redis_pttl_2})
        mock_redis_decr_or_set.side_effect = my_redis_decr_or_set
        client = DiscordClient(TEST_BOT_TOKEN, my_mock_redis)

        result = client.create_guild_role(
            guild_id=TEST_GUILD_ID, role_name=self.my_role_api['name']
        )
        self.assertEqual(result, self.my_role_obj)
        self.assertTrue(mock_sleep.called)

    def test_throw_exception_if_rate_limit_reached(
        self, mock_redis_decr_or_set, requests_mocker
    ):
        requests_mocker.post(
            f'{API_BASE_URL}guilds/{TEST_GUILD_ID}/roles', json=self.my_role_api
        )
        my_mock_redis = MagicMock(**{'pttl.side_effect': self.my_redis_pttl})
        mock_redis_decr_or_set.return_value = -1
        client = DiscordClient(TEST_BOT_TOKEN, my_mock_redis)
        try:
            client.create_guild_role(
                guild_id=TEST_GUILD_ID, role_name=self.my_role_api['name']
            )
        except Exception as ex:
            self.assertIsInstance(ex, DiscordRateLimitExhausted)
            self.assertEqual(ex.retry_after, TEST_RETRY_AFTER)

    @patch(MODULE_PATH + '.RATE_LIMIT_RETRIES', 1)
    @patch(MODULE_PATH + '.sleep', spec=True)
    def test_throw_exception_if_retries_are_exhausted(
        self, requests_mocker, mock_sleep, mock_redis_decr_or_set
    ):
        def my_redis_pttl_2(name: str):
            if name == DiscordClient._KEY_GLOBAL_BACKOFF_UNTIL:
                return -1
            return 100

        requests_mocker.post(
            f'{API_BASE_URL}guilds/{TEST_GUILD_ID}/roles', json=self.my_role_api
        )
        mock_sleep.side_effect = my_sleep
        my_mock_redis = MagicMock(**{'pttl.side_effect': my_redis_pttl_2})
        mock_redis_decr_or_set.return_value = -1
        client = DiscordClient(TEST_BOT_TOKEN, my_mock_redis)

        with self.assertRaises(RuntimeError):
            client.create_guild_role(
                guild_id=TEST_GUILD_ID, role_name=self.my_role_api['name']
            )

    def test_report_api_rate_limits(
        self, mock_redis_decr_or_set, requests_mocker
    ):
        headers = {
            'x-ratelimit-limit': '10',
            'x-ratelimit-remaining': '9',
            'x-ratelimit-reset-after': '10.000',
        }
        requests_mocker.post(
            f'{API_BASE_URL}guilds/{TEST_GUILD_ID}/roles',
            json=self.my_role_api,
            headers=headers
        )
        my_mock_redis = MagicMock(**{'pttl.side_effect': self.my_redis_pttl})
        mock_redis_decr_or_set.return_value = 5
        client = DiscordClient(TEST_BOT_TOKEN, my_mock_redis)
        result = client.create_guild_role(
            guild_id=TEST_GUILD_ID, role_name=self.my_role_api['name']
        )
        self.assertEqual(result, self.my_role_obj)

    def test_dont_report_api_rate_limits(
        self, mock_redis_decr_or_set, requests_mocker
    ):
        headers = {
            'x-ratelimit-limit': '10',
            'x-ratelimit-remaining': '5',
            'x-ratelimit-reset-after': '10.000',
        }
        requests_mocker.post(
            f'{API_BASE_URL}guilds/{TEST_GUILD_ID}/roles',
            json=self.my_role_api,
            headers=headers
        )
        my_mock_redis = MagicMock(**{'pttl.side_effect': self.my_redis_pttl})
        mock_redis_decr_or_set.return_value = 5
        client = DiscordClient(TEST_BOT_TOKEN, my_mock_redis)
        result = client.create_guild_role(
            guild_id=TEST_GUILD_ID, role_name=self.my_role_api['name']
        )
        self.assertEqual(result, self.my_role_obj)

    def test_ignore_errors_in_api_rate_limits(
        self, mock_redis_decr_or_set, requests_mocker
    ):
        headers = {
            'x-ratelimit-limit': '10',
            'x-ratelimit-remaining': '0',
            'x-ratelimit-reset-after': 'invalid',
        }
        requests_mocker.post(
            f'{API_BASE_URL}guilds/{TEST_GUILD_ID}/roles',
            json=self.my_role_api,
            headers=headers
        )
        my_mock_redis = MagicMock(**{'pttl.side_effect': self.my_redis_pttl})
        mock_redis_decr_or_set.return_value = 5
        client = DiscordClient(TEST_BOT_TOKEN, my_mock_redis)
        result = client.create_guild_role(
            guild_id=TEST_GUILD_ID, role_name=self.my_role_api['name']
        )
        self.assertEqual(result, self.my_role_obj)

    @patch(MODULE_PATH + '.DiscordClient._ensure_rate_limed_not_exhausted', spec=True)
    def test_can_turn_off_rate_limiting(
        self,
        requests_mocker,
        mock_ensure_rate_limed_not_exhausted,
        mock_redis_decr_or_set
    ):
        requests_mocker.post(
            f'{API_BASE_URL}guilds/{TEST_GUILD_ID}/roles', json=self.my_role_api
        )
        client = DiscordClient(TEST_BOT_TOKEN, mock_redis, is_rate_limited=False)
        result = client.create_guild_role(
            guild_id=TEST_GUILD_ID, role_name=self.my_role_api['name']
        )
        self.assertEqual(result, self.my_role_obj)
        self.assertFalse(mock_ensure_rate_limed_not_exhausted.called)


@patch(MODULE_PATH + '.DiscordClient._redis_decr_or_set', spec=True)
@requests_mock.Mocker()
class TestBackoffHandling(NoSocketsTestCase):

    my_role_api = create_discord_role_object(1, "alpha")
    my_role_obj = create_role(id=1, name="alpha")

    def test_dont_raise_exception_when_no_global_backoff(
        self, mock_redis_decr_or_set, requests_mocker
    ):
        requests_mocker.post(
            f'{API_BASE_URL}guilds/{TEST_GUILD_ID}/roles', json=self.my_role_api
        )
        my_mock_redis = MagicMock(**{'pttl.return_value': -1})
        mock_redis_decr_or_set.return_value = 5
        client = DiscordClient(TEST_BOT_TOKEN, my_mock_redis)
        result = client.create_guild_role(guild_id=TEST_GUILD_ID, role_name='dummy')
        self.assertEqual(result, self.my_role_obj)

    def test_raise_exception_when_global_backoff_in_effect(
        self, mock_redis_decr_or_set, requests_mocker
    ):
        requests_mocker.post(
            f'{API_BASE_URL}guilds/{TEST_GUILD_ID}/roles', json=self.my_role_api
        )
        retry_after = 1000
        my_mock_redis = MagicMock(**{'pttl.return_value': retry_after})
        mock_redis_decr_or_set.return_value = 5
        client = DiscordClient(TEST_BOT_TOKEN, my_mock_redis)
        try:
            client.create_guild_role(
                guild_id=TEST_GUILD_ID, role_name='dummy'
            )
        except Exception as ex:
            self.assertIsInstance(ex, DiscordTooManyRequestsError)
            self.assertEqual(ex.retry_after, retry_after)

    @patch(MODULE_PATH + '.sleep', spec=True)
    def test_just_wait_if_global_backoff_ends_soon(
        self, requests_mocker, mock_sleep, mock_redis_decr_or_set,
    ):
        requests_mocker.post(
            f'{API_BASE_URL}guilds/{TEST_GUILD_ID}/roles', json=self.my_role_api
        )
        retry_after = 50
        mock_sleep.side_effect = my_sleep
        my_mock_redis = MagicMock(**{'pttl.return_value': retry_after})
        mock_redis_decr_or_set.return_value = 5
        client = DiscordClient(TEST_BOT_TOKEN, my_mock_redis)
        client.create_guild_role(
            guild_id=TEST_GUILD_ID, role_name='dummy'
        )
        result = client.create_guild_role(guild_id=TEST_GUILD_ID, role_name='dummy')
        self.assertEqual(result, self.my_role_obj)
        self.assertTrue(mock_sleep.called)

    @patch(MODULE_PATH + '.DiscordClient._redis_set_if_longer', spec=True)
    def test_raise_exception_if_api_returns_429(
        self, requests_mocker, mock_redis_set_if_longer, mock_redis_decr_or_set,
    ):
        retry_after = 5000
        requests_mocker.post(
            f'{API_BASE_URL}guilds/{TEST_GUILD_ID}/roles',
            status_code=429,
            json={'retry_after': retry_after}
        )
        my_mock_redis = MagicMock(
            **{'pttl.side_effect': TestRateLimitMechanic.my_redis_pttl}
        )
        mock_redis_decr_or_set.return_value = 5

        client = DiscordClient(TEST_BOT_TOKEN, my_mock_redis)
        try:
            client.create_guild_role(
                guild_id=TEST_GUILD_ID, role_name='dummy'
            )
        except Exception as ex:
            self.assertIsInstance(ex, DiscordTooManyRequestsError)
            self.assertEqual(ex.retry_after, retry_after + DURATION_CONTINGENCY)
            self.assertTrue(mock_redis_set_if_longer.called)
            _, kwargs = mock_redis_set_if_longer.call_args
            self.assertEqual(kwargs['px'], retry_after + DURATION_CONTINGENCY)

    @patch(MODULE_PATH + '.DiscordClient._redis_set_if_longer', spec=True)
    def test_raise_exception_if_api_returns_429_no_retry_info(
        self, requests_mocker, mock_redis_set_if_longer, mock_redis_decr_or_set,
    ):
        requests_mocker.post(
            f'{API_BASE_URL}guilds/{TEST_GUILD_ID}/roles',
            status_code=429,
            json={}
        )
        my_mock_redis = MagicMock(
            **{'pttl.side_effect': TestRateLimitMechanic.my_redis_pttl}
        )
        mock_redis_decr_or_set.return_value = 5

        client = DiscordClient(TEST_BOT_TOKEN, my_mock_redis)
        try:
            client.create_guild_role(
                guild_id=TEST_GUILD_ID, role_name='dummy'
            )
        except Exception as ex:
            self.assertIsInstance(ex, DiscordTooManyRequestsError)
            self.assertEqual(ex.retry_after, DEFAULT_BACKOFF_DELAY)
            self.assertTrue(mock_redis_set_if_longer.called)
            _, kwargs = mock_redis_set_if_longer.call_args
            self.assertEqual(kwargs['px'], DEFAULT_BACKOFF_DELAY)

    @patch(MODULE_PATH + '.DiscordClient._redis_set_if_longer', spec=True)
    def test_raise_exception_if_api_returns_429_ignore_value_error(
        self, requests_mocker, mock_redis_set_if_longer, mock_redis_decr_or_set,
    ):
        requests_mocker.post(
            f'{API_BASE_URL}guilds/{TEST_GUILD_ID}/roles',
            status_code=429,
            json={'retry_after': "invalid"}
        )
        my_mock_redis = MagicMock(
            **{'pttl.side_effect': TestRateLimitMechanic.my_redis_pttl}
        )
        mock_redis_decr_or_set.return_value = 5

        client = DiscordClient(TEST_BOT_TOKEN, my_mock_redis)
        try:
            client.create_guild_role(
                guild_id=TEST_GUILD_ID, role_name='dummy'
            )
        except Exception as ex:
            self.assertIsInstance(ex, DiscordTooManyRequestsError)
            self.assertEqual(ex.retry_after, DEFAULT_BACKOFF_DELAY)
            self.assertTrue(mock_redis_set_if_longer.called)
            _, kwargs = mock_redis_set_if_longer.call_args
            self.assertEqual(kwargs['px'], DEFAULT_BACKOFF_DELAY)


class TestRedisDecode(NoSocketsTestCase):

    def test_decode_string(self):
        self.assertEqual(
            DiscordClient._redis_decode(b'MyTest123'), 'MyTest123'
        )

    def test_decode_bool(self):
        self.assertTrue(DiscordClient._redis_decode(True))
        self.assertFalse(DiscordClient._redis_decode(False))

    def test_decode_none(self):
        self.assertIsNone(DiscordClient._redis_decode(None))


class TestTouchLuaScripts(NoSocketsTestCase):

    def test__redis_script_decr_or_set(self):
        client = DiscordClient(TEST_BOT_TOKEN, mock_redis)
        client._redis_decr_or_set(name='dummy', value=5, px=1000)

    def test_redis_set_if_longer(self):
        client = DiscordClient(TEST_BOT_TOKEN, mock_redis)
        client._redis_set_if_longer(name='dummy', value=5, px=1000)



from django.db import models
from django.core.exceptions import ValidationError
from django.utils.translation import gettext_lazy as _

from uuid import uuid4


class AnalyticsIdentifier(models.Model):

    identifier = models.UUIDField(
        default=uuid4,
        editable=False)

    def save(self, *args, **kwargs):
        if not self.pk and AnalyticsIdentifier.objects.exists():
            # Force a single object
            raise ValidationError('There is can be only one \
                                    AnalyticsIdentifier instance')
        self.pk = self.id = 1  # If this happens to be deleted and recreated, force it to be 1
        return super().save(*args, **kwargs)


class AnalyticsTokens(models.Model):

    class Analytics_Type(models.TextChoices):
        GA_U = 'GA-U', _('Google Analytics Universal')
        GA_V4 = 'GA-V4', _('Google Analytics V4')

    name = models.CharField(max_length=254)
    type = models.CharField(max_length=254, choices=Analytics_Type.choices)
    token = models.CharField(max_length=254, blank=False)
    secret = models.CharField(max_length=254, blank=True)
    send_stats = models.BooleanField(default=False)



# Generated by Django 1.10.2 on 2016-12-12 01:11

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('auth', '0008_alter_user_username_max_length'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='AuthTS',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('auth_group', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='auth.Group')),
            ],
            options={
                'db_table': 'teamspeak3_authts',
                'verbose_name': 'Auth / TS Group',
            },
        ),
        migrations.CreateModel(
            name='TSgroup',
            fields=[
                ('ts_group_id', models.IntegerField(primary_key=True, serialize=False)),
                ('ts_group_name', models.CharField(max_length=30)),
            ],
            options={
                'db_table': 'teamspeak3_tsgroup',
                'verbose_name': 'TS Group',
            },
        ),
        migrations.CreateModel(
            name='UserTSgroup',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('ts_group', models.ManyToManyField(to='teamspeak3.TSgroup')),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'db_table': 'teamspeak3_usertsgroup',
                'verbose_name': 'User TS Group',
            },
        ),
        migrations.AddField(
            model_name='authts',
            name='ts_group',
            field=models.ManyToManyField(to='teamspeak3.TSgroup'),
        ),
    ]



# Generated by Django 3.2.8 on 2021-10-26 09:19

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('authentication', '0018_alter_state_name_length'),
        ('authentication', '0018_state_member_factions'),
    ]

    operations = [
    ]



from typing import List

from django.contrib.auth.models import User, Group, Permission
from django.db.models.signals import m2m_changed, pre_save, post_save
from django.test import TestCase

from esi.models import Token

from allianceauth.authentication.models import (
    UserProfile, State, get_guest_state
)
from allianceauth.eveonline.models import EveCharacter
from allianceauth.authentication.signals import (
    state_member_alliances_changed,
    state_member_characters_changed,
    state_member_corporations_changed,
    state_saved,
    reassess_on_profile_save,
    assign_state_on_active_change,
    check_state_on_character_update
)
from allianceauth.services.signals import (
    m2m_changed_group_permissions,
    m2m_changed_user_permissions,
    m2m_changed_state_permissions,
    m2m_changed_user_groups, disable_services_on_inactive,
    process_main_character_change,
    process_main_character_update
)


class AuthUtils:
    """Utilities for making it easier to create tests for Alliance Auth"""

    @staticmethod
    def _create_user(username):
        return User.objects.create(username=username)

    @classmethod
    def create_user(cls, username, disconnect_signals=False):
        """create a new user

        username: Name of the user

        disconnect_signals: whether to run process without signals
        """
        if disconnect_signals:
            cls.disconnect_signals()
        user = cls._create_user(username)
        if disconnect_signals:
            cls.connect_signals()
        return user

    @staticmethod
    def _create_state(name, priority, member_characters=None, member_corporations=None, member_alliances=None, public=False):
        state = State.objects.create(name=name, priority=priority)
        if member_characters:
            state.member_characters.add(member_characters)
        if member_corporations:
            state.member_corporations.add(member_corporations)
        if member_alliances:
            state.member_alliances.add(member_alliances)
        return state

    @classmethod
    def create_state(cls, name, priority, member_characters=None, member_corporations=None, member_alliances=None, public=False, disconnect_signals=False):
        if disconnect_signals:
            cls.disconnect_signals()
        state = cls._create_state(name, priority, member_characters=member_characters, member_corporations=member_corporations, public=public, member_alliances=member_alliances)
        if disconnect_signals:
            cls.connect_signals()
        return state

    @classmethod
    def get_member_state(cls):
        try:
            return State.objects.get(name='Member')
        except State.DoesNotExist:
            return cls.create_state('Member', 100, disconnect_signals=True)

    @classmethod
    def get_guest_state(cls):
        cls.disconnect_signals()
        state = get_guest_state()
        cls.connect_signals()
        return state

    @classmethod
    def assign_state(cls, user, state, disconnect_signals=False):
        if disconnect_signals:
            cls.disconnect_signals()
        profile = user.profile
        profile.state = state
        profile.save()
        if disconnect_signals:
            cls.connect_signals()

    @classmethod
    def create_member(cls, username):
        user = cls.create_user(username, disconnect_signals=True)
        state = cls.get_member_state()
        cls.assign_state(user, state, disconnect_signals=True)
        cls.disconnect_signals()
        g = Group.objects.get_or_create(name='Member')[0]
        user.groups.add(g)
        cls.connect_signals()
        return user

    @classmethod
    def disconnect_signals(cls):
        m2m_changed.disconnect(m2m_changed_user_groups, sender=User.groups.through)
        m2m_changed.disconnect(m2m_changed_group_permissions, sender=Group.permissions.through)
        m2m_changed.disconnect(m2m_changed_user_permissions, sender=User.user_permissions.through)
        m2m_changed.disconnect(m2m_changed_state_permissions, sender=State.permissions.through)
        pre_save.disconnect(disable_services_on_inactive, sender=User)
        m2m_changed.disconnect(state_member_corporations_changed, sender=State.member_corporations.through)
        m2m_changed.disconnect(state_member_characters_changed, sender=State.member_characters.through)
        m2m_changed.disconnect(state_member_alliances_changed, sender=State.member_alliances.through)
        post_save.disconnect(state_saved, sender=State)
        post_save.disconnect(reassess_on_profile_save, sender=UserProfile)
        pre_save.disconnect(assign_state_on_active_change, sender=User)
        pre_save.disconnect(process_main_character_change, sender=UserProfile)
        pre_save.disconnect(process_main_character_update, sender=EveCharacter)
        post_save.disconnect(
            check_state_on_character_update, sender=EveCharacter
        )

    @classmethod
    def connect_signals(cls):
        m2m_changed.connect(m2m_changed_user_groups, sender=User.groups.through)
        m2m_changed.connect(m2m_changed_group_permissions, sender=Group.permissions.through)
        m2m_changed.connect(m2m_changed_user_permissions, sender=User.user_permissions.through)
        m2m_changed.connect(m2m_changed_state_permissions, sender=State.permissions.through)
        pre_save.connect(disable_services_on_inactive, sender=User)
        m2m_changed.connect(state_member_corporations_changed, sender=State.member_corporations.through)
        m2m_changed.connect(state_member_characters_changed, sender=State.member_characters.through)
        m2m_changed.connect(state_member_alliances_changed, sender=State.member_alliances.through)
        post_save.connect(state_saved, sender=State)
        post_save.connect(reassess_on_profile_save, sender=UserProfile)
        pre_save.connect(assign_state_on_active_change, sender=User)
        pre_save.connect(process_main_character_change, sender=UserProfile)
        pre_save.connect(process_main_character_update, sender=EveCharacter)
        post_save.connect(check_state_on_character_update, sender=EveCharacter)

    @classmethod
    def add_main_character(cls, user, name, character_id, corp_id=2345, corp_name='', corp_ticker='', alliance_id=None,
                            alliance_name='', faction_id=None, faction_name=''):
        if alliance_id:
            try:
                alliance_id = int(alliance_id)
            except Exception:
                alliance_id = None

        char = EveCharacter.objects.create(
            character_id=int(character_id),
            character_name=name,
            corporation_id=int(corp_id),
            corporation_name=corp_name,
            corporation_ticker=corp_ticker,
            alliance_id=alliance_id,
            alliance_name=alliance_name,
            faction_id=faction_id,
            faction_name=faction_name
        )
        UserProfile.objects.update_or_create(user=user, defaults={'main_character': char})

    @classmethod
    def add_main_character_2(
        cls,
        user,
        name,
        character_id,
        corp_id=2345,
        corp_name='',
        corp_ticker='',
        alliance_id=None,
        alliance_name='',
        disconnect_signals=False
    ) -> EveCharacter:
        """new version that works in all cases"""
        if disconnect_signals:
            cls.disconnect_signals()

        if alliance_id:
            try:
                alliance_id = int(alliance_id)
            except Exception:
                alliance_id = None

        char = EveCharacter.objects.create(
            character_id=int(character_id),
            character_name=name,
            corporation_id=int(corp_id),
            corporation_name=corp_name,
            corporation_ticker=corp_ticker,
            alliance_id=alliance_id,
            alliance_name=alliance_name,
        )
        user.profile.main_character = char
        user.profile.save()

        if disconnect_signals:
            cls.connect_signals()

        return char

    @classmethod
    def add_permissions_to_groups(cls, perms, groups, disconnect_signals=True):
        if disconnect_signals:
            cls.disconnect_signals()

        for group in groups:
            for perm in perms:
                group.permissions.add(perm)
            group = Group.objects.get(pk=group.pk)    # reload permission cache

        if disconnect_signals:
            cls.connect_signals()

    @classmethod
    def add_permissions_to_state(cls, perms, states, disconnect_signals=True):
        return cls.add_permissions_to_groups(
            perms, states, disconnect_signals=disconnect_signals
        )

    @classmethod
    def add_permissions_to_user(cls, perms, user, disconnect_signals=True) -> User:
        """add list of permissions to user

        perms: list of Permission objects

        user: user object

        disconnect_signals: whether to run process without signals
        """
        if disconnect_signals:
            cls.disconnect_signals()

        for perm in perms:
            user.user_permissions.add(perm)

        user = User.objects.get(pk=user.pk)     # reload permission cache
        if disconnect_signals:
            cls.connect_signals()

        return user

    @classmethod
    def add_permission_to_user_by_name(
        cls, perm, user, disconnect_signals=True
    ) -> User:
        """returns permission specified by qualified name

        perm: Permission name as 'app_label.codename'

        user: user object

        disconnect_signals: whether to run process without signals
        """
        p = cls.get_permission_by_name(perm)
        return cls.add_permissions_to_user([p], user, disconnect_signals)

    @classmethod
    def add_permissions_to_user_by_name(
        cls, perms: List[str], user: User, disconnect_signals: bool = True
    ) -> User:
        """Add permissions given by name to a user

        Args:
            perms: List of permission names as 'app_label.codename'
            user: user object
            disconnect_signals: whether to run process without signals

        Returns:
            Updated user object
        """
        permissions = [cls.get_permission_by_name(perm) for perm in perms]
        return cls.add_permissions_to_user(permissions, user, disconnect_signals)

    @staticmethod
    def get_permission_by_name(perm: str) -> Permission:
        """returns permission specified by qualified name

        perm: Permission name as 'app_label.codename'

        Returns: Permission object or throws exception if not found
        """
        perm_parts = perm.split('.')
        if len(perm_parts) != 2:
            raise ValueError('Invalid format for permission name')

        return Permission.objects.get(
            content_type__app_label=perm_parts[0], codename=perm_parts[1]
        )


class BaseViewTestCase(TestCase):
    def setUp(self):
        self.member = AuthUtils.create_member('auth_member')
        self.member.email = 'auth_member@example.com'
        self.member.save()
        AuthUtils.add_main_character(self.member, 'auth_member', '12345', corp_id='111', corp_name='Test Corporation',
                                    corp_ticker='TESTR')

    def login(self):
        token = Token.objects.create(character_id='12345', character_name='auth_member', character_owner_hash='1', user=self.member, access_token='1')
        self.client.login(token=token)



# Generated by Django 2.2.9 on 2020-01-30 13:14

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('groupmanagement', '0012_group_leads'),
    ]

    operations = [
        migrations.AlterField(
            model_name='requestlog',
            name='date',
            field=models.DateTimeField(auto_now_add=True),
        ),
    ]



"""Helpers for Task Statistics."""

import logging

from redis import Redis, RedisError

from allianceauth.utils.cache import get_redis_client

logger = logging.getLogger(__name__)


class _RedisStub:
    """Stub of a Redis client.

    It's purpose is to prevent EventSeries objects from trying to access Redis
    when it is not available. e.g. when the Sphinx docs are rendered by readthedocs.org.
    """

    IS_STUB = True

    def delete(self, *args, **kwargs):
        pass

    def incr(self, *args, **kwargs):
        return 0

    def zadd(self, *args, **kwargs):
        pass

    def zcount(self, *args, **kwargs):
        pass

    def zrangebyscore(self, *args, **kwargs):
        pass


def get_redis_client_or_stub() -> Redis:
    """Return AA's default cache client or a stub if Redis is not available."""
    redis = get_redis_client()
    try:
        if not redis.ping():
            raise RuntimeError()
    except (AttributeError, RedisError, RuntimeError):
        logger.exception(
            "Failed to establish a connection with Redis. "
            "This EventSeries object is disabled.",
        )
        return _RedisStub()
    return redis



from allianceauth import hooks
from allianceauth.theme.hooks import ThemeHook


class BootstrapThemeHook(ThemeHook):
    """
    Bootstrap in all its glory!
    https://getbootstrap.com/
    """

    def __init__(self):
        ThemeHook.__init__(
            self,
            "Bootstrap",
            "Powerful, extensible, and feature-packed frontend toolkit.",
            css=[{
                "url": "https://cdnjs.cloudflare.com/ajax/libs/bootstrap/5.3.0/css/bootstrap.min.css",
                "integrity": "sha512-t4GWSVZO1eC8BM339Xd7Uphw5s17a86tIZIj8qRxhnKub6WoyhnrxeCIMeAqBPgdZGlCcG2PrZjMc+Wr78+5Xg=="
            }],
            js=[{
                "url": "https://cdnjs.cloudflare.com/ajax/libs/popper.js/2.11.8/umd/popper.min.js",
                "integrity": "sha512-TPh2Oxlg1zp+kz3nFA0C5vVC6leG/6mm1z9+mA81MI5eaUVqasPLO8Cuk4gMF4gUfP5etR73rgU/8PNMsSesoQ=="
            }, {
                "url": "https://cdnjs.cloudflare.com/ajax/libs/bootstrap/5.3.0/js/bootstrap.min.js",
                "integrity": "sha512-3dZ9wIrMMij8rOH7X3kLfXAzwtcHpuYpEgQg1OA4QAob1e81H8ntUQmQm3pBudqIoySO5j0tHN4ENzA6+n2r4w=="
            }],
            header_padding="3.5em"
        )


@hooks.register('theme_hook')
def register_darkly_hook():
    return BootstrapThemeHook()



# Generated by Django 4.0.7 on 2022-08-14 16:23

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('eveonline', '0016_character_names_are_not_unique'),
    ]

    operations = [
        migrations.AlterField(
            model_name='eveallianceinfo',
            name='alliance_name',
            field=models.CharField(max_length=254, db_index=True),
        ),
        migrations.AlterField(
            model_name='evecorporationinfo',
            name='corporation_name',
            field=models.CharField(max_length=254, db_index=True),
        ),
    ]



from copy import copy
from typing import Iterable, List, Optional, Set, Tuple

from .models import Role


class RolesSet:
    """Container of Discord roles with added functionality.

    Objects of this class are immutable and work in many ways like sets.

    Ideally objects are initialized from raw API responses,
    e.g. from DiscordClient.guild.roles().

    Args:
        roles_lst: List of dicts, each defining a role
    """
    def __init__(self, roles_lst: Iterable[Role]) -> None:
        if not isinstance(roles_lst, (list, set, tuple)):
            raise TypeError('roles_lst must be of type list, set or tuple')
        self._roles = dict()
        self._roles_by_name = dict()
        for role in list(roles_lst):
            if not isinstance(role, Role):
                raise TypeError('Roles must be of type Role: %s' % role)
            self._roles[role.id] = role
            self._roles_by_name[role.name] = role

    def __repr__(self) -> str:
        if self._roles_by_name:
            roles = '"' + '", "'.join(sorted(list(self._roles_by_name.keys()))) + '"'
        else:
            roles = ""
        return f'{self.__class__.__name__}([{roles}])'

    def __eq__(self, other):
        if isinstance(other, type(self)):
            return self.ids() == other.ids()
        return NotImplemented

    def __hash__(self):
        return hash(tuple(sorted(self._roles.keys())))

    def __iter__(self):
        yield from self._roles.values()

    def __contains__(self, item) -> bool:
        return int(item) in self._roles

    def __len__(self):
        return len(self._roles.keys())

    def has_roles(self, role_ids: Set[int]) -> bool:
        """True if this objects contains all roles defined by given role_ids
        incl. managed roles.
        """
        role_ids = {int(id) for id in role_ids}
        all_role_ids = self._roles.keys()
        return role_ids.issubset(all_role_ids)

    def ids(self) -> Set[int]:
        """Set of all role IDs."""
        return set(self._roles.keys())

    def subset(
        self,
        role_ids: Iterable[int] = None,
        managed_only: bool = False,
        role_names: Iterable[str] = None
    ) -> "RolesSet":
        """Create instance containing the subset of roles

        Args:
            role_ids: role ids must be in the provided list
            managed_only: roles must be managed
            role_names: role names must match provided list (not case sensitive)
        """
        if role_ids is not None:
            role_ids = {int(id) for id in role_ids}

        if role_ids is not None and not managed_only:
            return type(self)([
                role for role_id, role in self._roles.items() if role_id in role_ids
            ])

        elif role_ids is None and managed_only:
            return type(self)([
                role for _, role in self._roles.items() if role.managed
            ])

        elif role_ids is not None and managed_only:
            return type(self)([
                role for role_id, role in self._roles.items()
                if role_id in role_ids and role.managed
            ])

        elif role_ids is None and managed_only is False and role_names is not None:
            role_names = {Role.sanitize_name(name).lower() for name in role_names}
            return type(self)([
                role for role in self._roles.values()
                if role.name.lower() in role_names
            ])

        return copy(self)

    def union(self, other: object) -> "RolesSet":
        """Create instance that is the union of this roles object with other."""
        return type(self)(list(self) + list(other))

    def difference(self, other: object) -> "RolesSet":
        """Create instance that only contains the roles
        that exist in the current objects, but not in other.
        """
        new_ids = self.ids().difference(other.ids())
        return self.subset(role_ids=new_ids)

    def role_by_name(self, role_name: str) -> Optional[Role]:
        """Role if one with matching name is found else None."""
        role_name = Role.sanitize_name(role_name)
        if role_name in self._roles_by_name:
            return self._roles_by_name[role_name]
        return None

    @classmethod
    def create_from_matched_roles(
        cls, matched_roles: List[Tuple[Role, bool]]
    ) -> "RolesSet":
        """Create new instance from the given list of matches roles.

        Args:
            matches_roles: list of matches roles
        """
        raw_roles = [x[0] for x in matched_roles]
        return cls(raw_roles)



from unittest.mock import patch

from django.test import TestCase

from allianceauth.utils.django import StartupCommand

MODULE_PATH = "allianceauth.utils.django"


class TestStartupCommand(TestCase):
    def test_should_detect_management_command(self):
        # when
        with patch(MODULE_PATH + ".sys") as m:
            m.argv = ["manage.py", "check"]
            info = StartupCommand()
        # then
        self.assertTrue(info.is_management_command)

    def test_should_detect_not_a_management_command(self):
        # when
        with patch(MODULE_PATH + ".sys") as m:
            m.argv = ['/home/python/allianceauth-dev/venv/bin/gunicorn', 'myauth.wsgi']
            info = StartupCommand()
        # then
        self.assertFalse(info.is_management_command)



# Generated by Django 3.1.13 on 2021-10-12 01:04

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('eveonline', '0014_auto_20210105_1413'),
    ]

    operations = [
        migrations.CreateModel(
            name='EveFactionInfo',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('faction_id', models.PositiveIntegerField(db_index=True, unique=True)),
                ('faction_name', models.CharField(max_length=254, unique=True)),
            ],
        ),
        migrations.AddField(
            model_name='evecharacter',
            name='faction_id',
            field=models.PositiveIntegerField(blank=True, default=None, null=True),
        ),
        migrations.AddField(
            model_name='evecharacter',
            name='faction_name',
            field=models.CharField(blank=True, default='', max_length=254, null=True),
        ),
        migrations.AddIndex(
            model_name='evecharacter',
            index=models.Index(fields=['faction_id'], name='eveonline_e_faction_d5274e_idx'),
        ),
    ]



# Generated by Django 1.10.1 on 2017-01-07 06:47

from django.db import migrations

def count_completed_fields(model):
    return len([True for key, value in model.__dict__.items() if bool(value)])

def forward(apps, schema_editor):
    # this ensures only one model exists per user
    AuthServicesInfo = apps.get_model('authentication', 'AuthServicesInfo')
    users = {a.user for a in AuthServicesInfo.objects.all()}
    for u in users:
        auths = AuthServicesInfo.objects.filter(user=u)
        if auths.count() > 1:
            pk = auths[0].pk
            largest = 0
            for auth in auths:
                completed = count_completed_fields(auth)
                if completed > largest:
                    largest = completed
                    pk = auth.pk
            auths.exclude(pk=pk).delete()

    # ensure all users have a model
    User = apps.get_model('auth', 'User')
    for u in User.objects.exclude(pk__in=[user.pk for user in users]):
        AuthServicesInfo.objects.create(user=u)


class Migration(migrations.Migration):

    dependencies = [
        ('authentication', '0009_auto_20161021_0228'),
    ]

    operations = [
        migrations.RunPython(forward, migrations.RunPython.noop)
    ]



# Generated by Django 1.10.5 on 2017-02-02 05:59

from django.db import migrations
from django.conf import settings
from django.core.exceptions import ObjectDoesNotExist
from django.contrib.auth.management import create_permissions

import logging

logger = logging.getLogger(__name__)


def migrate_service_enabled(apps, schema_editor):
    for app_config in apps.get_app_configs():
        app_config.models_module = True
        create_permissions(app_config, apps=apps, verbosity=0)
        app_config.models_module = None

    Group = apps.get_model("auth", "Group")
    Permission = apps.get_model("auth", "Permission")
    DiscordUser = apps.get_model("discord", "DiscordUser")

    perm = Permission.objects.get(codename='access_discord')

    member_group_name = getattr(settings, 'DEFAULT_AUTH_GROUP', 'Member')
    blue_group_name = getattr(settings, 'DEFAULT_BLUE_GROUP', 'Blue')

    # Migrate members
    if DiscordUser.objects.filter(user__groups__name=member_group_name).exists() or \
            getattr(settings, 'ENABLE_AUTH_DISCORD', False):
        try:
            group = Group.objects.get(name=member_group_name)
            group.permissions.add(perm)
        except ObjectDoesNotExist:
            logger.warning('Failed to migrate ENABLE_AUTH_DISCORD setting')

    # Migrate blues
    if DiscordUser.objects.filter(user__groups__name=blue_group_name).exists() or \
            getattr(settings, 'ENABLE_BLUE_DISCORD', False):
        try:
            group = Group.objects.get(name=blue_group_name)
            group.permissions.add(perm)
        except ObjectDoesNotExist:
            logger.warning('Failed to migrate ENABLE_BLUE_DISCORD setting')


class Migration(migrations.Migration):

    dependencies = [
        ('discord', '0001_initial'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='discorduser',
            options={'permissions': (('access_discord', 'Can access the Discord service'),)},
        ),
        migrations.RunPython(migrate_service_enabled, migrations.RunPython.noop),
    ]



import random
import string
import calendar
import re
from datetime import datetime

from passlib.apps import phpbb3_context
from django.db import connections
from allianceauth.eveonline.models import EveCharacter

import logging

from django.conf import settings

logger = logging.getLogger(__name__)


TABLE_PREFIX = getattr(settings, 'PHPBB3_TABLE_PREFIX', 'phpbb_')


class Phpbb3Manager:
    SQL_ADD_USER =  r"INSERT INTO %susers (username, username_clean, " \
                    r"user_password, user_email, group_id, user_regdate, user_permissions, " \
                    r"user_sig, user_lang) VALUES (%%s, %%s, %%s, %%s, %%s, %%s, %%s, %%s, 'en')" % TABLE_PREFIX

    SQL_DEL_USER = r"DELETE FROM %susers where username = %%s" % TABLE_PREFIX

    SQL_DIS_USER = r"UPDATE %susers SET user_email= %%s, user_password=%%s WHERE username = %%s" % TABLE_PREFIX

    SQL_USER_ID_FROM_USERNAME = r"SELECT user_id from %susers WHERE username = %%s" % TABLE_PREFIX

    SQL_ADD_USER_GROUP = r"INSERT INTO %suser_group (group_id, user_id, user_pending) VALUES (%%s, %%s, %%s)" % TABLE_PREFIX

    SQL_GET_GROUP_ID = r"SELECT group_id from %sgroups WHERE group_name = %%s" % TABLE_PREFIX

    SQL_ADD_GROUP = r"INSERT INTO %sgroups (group_name,group_desc,group_legend) VALUES (%%s,%%s,0)" % TABLE_PREFIX

    SQL_UPDATE_USER_PASSWORD = r"UPDATE %susers SET user_password = %%s WHERE username = %%s" % TABLE_PREFIX

    SQL_REMOVE_USER_GROUP = r"DELETE FROM %suser_group WHERE user_id=%%s AND group_id=%%s " % TABLE_PREFIX

    SQL_GET_ALL_GROUPS = r"SELECT group_id, group_name FROM %sgroups" % TABLE_PREFIX

    SQL_GET_USER_GROUPS = r"SELECT %(prefix)sgroups.group_name FROM %(prefix)sgroups , %(prefix)suser_group WHERE " \
                        r"%(prefix)suser_group.group_id = %(prefix)sgroups.group_id AND user_id=%%s" % {'prefix': TABLE_PREFIX}

    SQL_ADD_USER_AVATAR = r"UPDATE %susers SET user_avatar_type=2, user_avatar_width=64, user_avatar_height=64, " \
                        "user_avatar=%%s WHERE user_id = %%s" % TABLE_PREFIX

    SQL_CLEAR_USER_PERMISSIONS = r"UPDATE %susers SET user_permissions = '' WHERE user_id = %%s" % TABLE_PREFIX

    SQL_DEL_SESSION = r"DELETE FROM %ssessions where session_user_id = %%s" % TABLE_PREFIX

    SQL_DEL_AUTOLOGIN = r"DELETE FROM %ssessions_keys where user_id = %%s" % TABLE_PREFIX

    def __init__(self):
        pass

    @staticmethod
    def __add_avatar(username, characterid):
        logger.debug(f"Adding EVE character id {characterid} portrait as phpbb avater for user {username}")
        avatar_url = EveCharacter.generic_portrait_url(characterid, 64)
        cursor = connections['phpbb3'].cursor()
        userid = Phpbb3Manager.__get_user_id(username)
        cursor.execute(Phpbb3Manager.SQL_ADD_USER_AVATAR, [avatar_url, userid])

    @staticmethod
    def __generate_random_pass():
        return ''.join([random.choice(string.ascii_letters + string.digits) for n in range(16)])

    @staticmethod
    def __gen_hash(password):
        return phpbb3_context.encrypt(password)

    @staticmethod
    def __santatize_username(username):
        sanatized = username.replace(" ", "_")
        sanatized = sanatized.replace("'", "_")
        return sanatized.lower()

    @staticmethod
    def _sanitize_groupname(name):
        name = name.strip(' _')
        return re.sub(r'[^\w.-]', '', name)

    @staticmethod
    def __get_group_id(groupname):
        logger.debug("Getting phpbb3 group id for groupname %s" % groupname)
        cursor = connections['phpbb3'].cursor()
        cursor.execute(Phpbb3Manager.SQL_GET_GROUP_ID, [groupname])
        row = cursor.fetchone()
        logger.debug(f"Got phpbb group id {row[0]} for groupname {groupname}")
        return row[0]

    @staticmethod
    def __get_user_id(username):
        logger.debug("Getting phpbb3 user id for username %s" % username)
        cursor = connections['phpbb3'].cursor()
        cursor.execute(Phpbb3Manager.SQL_USER_ID_FROM_USERNAME, [username])
        row = cursor.fetchone()
        if row is not None:
            logger.debug(f"Got phpbb user id {row[0]} for username {username}")
            return row[0]
        else:
            logger.error("Username %s not found on phpbb. Unable to determine user id." % username)
            return None

    @staticmethod
    def __get_all_groups():
        logger.debug("Getting all phpbb3 groups.")
        cursor = connections['phpbb3'].cursor()
        cursor.execute(Phpbb3Manager.SQL_GET_ALL_GROUPS)
        rows = cursor.fetchall()
        out = {}
        for row in rows:
            out[row[1]] = row[0]
        logger.debug("Got phpbb groups %s" % out)
        return out

    @staticmethod
    def __get_user_groups(userid):
        logger.debug("Getting phpbb3 user id %s groups" % userid)
        cursor = connections['phpbb3'].cursor()
        cursor.execute(Phpbb3Manager.SQL_GET_USER_GROUPS, [userid])
        out = [row[0] for row in cursor.fetchall()]
        logger.debug(f"Got user {userid} phpbb groups {out}")
        return out

    @staticmethod
    def __get_current_utc_date():
        d = datetime.utcnow()
        unixtime = calendar.timegm(d.utctimetuple())
        return unixtime

    @staticmethod
    def __create_group(groupname):
        logger.debug("Creating phpbb3 group %s" % groupname)
        cursor = connections['phpbb3'].cursor()
        cursor.execute(Phpbb3Manager.SQL_ADD_GROUP, [groupname, groupname])
        logger.info("Created phpbb group %s" % groupname)
        return Phpbb3Manager.__get_group_id(groupname)

    @staticmethod
    def __add_user_to_group(userid, groupid):
        logger.debug(f"Adding phpbb3 user id {userid} to group id {groupid}")
        try:
            cursor = connections['phpbb3'].cursor()
            cursor.execute(Phpbb3Manager.SQL_ADD_USER_GROUP, [groupid, userid, 0])
            cursor.execute(Phpbb3Manager.SQL_CLEAR_USER_PERMISSIONS, [userid])
            logger.info(f"Added phpbb user id {userid} to group id {groupid}")
        except:
            logger.exception(f"Unable to add phpbb user id {userid} to group id {groupid}")
            pass

    @staticmethod
    def __remove_user_from_group(userid, groupid):
        logger.debug(f"Removing phpbb3 user id {userid} from group id {groupid}")
        try:
            cursor = connections['phpbb3'].cursor()
            cursor.execute(Phpbb3Manager.SQL_REMOVE_USER_GROUP, [userid, groupid])
            cursor.execute(Phpbb3Manager.SQL_CLEAR_USER_PERMISSIONS, [userid])
            logger.info(f"Removed phpbb user id {userid} from group id {groupid}")
        except:
            logger.exception(f"Unable to remove phpbb user id {userid} from group id {groupid}")
            pass

    @staticmethod
    def add_user(username, email, groups, characterid):
        logger.debug("Adding phpbb user with username {}, email {}, groups {}, characterid {}".format(
            username, email, groups, characterid))
        cursor = connections['phpbb3'].cursor()

        username_clean = Phpbb3Manager.__santatize_username(username)
        password = Phpbb3Manager.__generate_random_pass()
        pwhash = Phpbb3Manager.__gen_hash(password)
        logger.debug(f"Proceeding to add phpbb user {username_clean} and pwhash starting with {pwhash[0:5]}")
        # check if the username was simply revoked
        if Phpbb3Manager.check_user(username_clean):
            logger.warning("Unable to add phpbb user with username %s - already exists. Updating user instead." % username)
            Phpbb3Manager.__update_user_info(username_clean, email, pwhash)
        else:
            try:

                cursor.execute(Phpbb3Manager.SQL_ADD_USER, [username_clean, username_clean, pwhash,
                                                            email, 2, Phpbb3Manager.__get_current_utc_date(),
                                                            "", ""])
                Phpbb3Manager.update_groups(username_clean, groups)
                Phpbb3Manager.__add_avatar(username_clean, characterid)
                logger.info("Added phpbb user %s" % username_clean)
            except:
                logger.exception("Unable to add phpbb user %s" % username_clean)
                pass

        return username_clean, password

    @staticmethod
    def disable_user(username):
        logger.debug("Disabling phpbb user %s" % username)
        cursor = connections['phpbb3'].cursor()

        password = Phpbb3Manager.__gen_hash(Phpbb3Manager.__generate_random_pass())
        revoke_email = "revoked@localhost"
        try:
            pwhash = Phpbb3Manager.__gen_hash(password)
            cursor.execute(Phpbb3Manager.SQL_DIS_USER, [revoke_email, pwhash, username])
            userid = Phpbb3Manager.__get_user_id(username)
            cursor.execute(Phpbb3Manager.SQL_DEL_AUTOLOGIN, [userid])
            cursor.execute(Phpbb3Manager.SQL_DEL_SESSION, [userid])
            Phpbb3Manager.update_groups(username, [])
            logger.info("Disabled phpbb user %s" % username)
            return True
        except TypeError:
            logger.exception("TypeError occured while disabling user %s - failed to disable." % username)
            return False

    @staticmethod
    def delete_user(username):
        logger.debug("Deleting phpbb user %s" % username)
        cursor = connections['phpbb3'].cursor()

        if Phpbb3Manager.check_user(username):
            cursor.execute(Phpbb3Manager.SQL_DEL_USER, [username])
            logger.info("Deleted phpbb user %s" % username)
            return True
        logger.error("Unable to delete phpbb user %s - user not found on phpbb." % username)
        return False

    @staticmethod
    def update_groups(username, groups):
        userid = Phpbb3Manager.__get_user_id(username)
        logger.debug(f"Updating phpbb user {username} with id {userid} groups {groups}")
        if userid is not None:
            forum_groups = Phpbb3Manager.__get_all_groups()
            user_groups = set(Phpbb3Manager.__get_user_groups(userid))
            act_groups = {Phpbb3Manager._sanitize_groupname(g) for g in groups}
            addgroups = act_groups - user_groups
            remgroups = user_groups - act_groups
            logger.info(f"Updating phpbb user {username} groups - adding {addgroups}, removing {remgroups}")
            for g in addgroups:
                if not g in forum_groups:
                    forum_groups[g] = Phpbb3Manager.__create_group(g)
                Phpbb3Manager.__add_user_to_group(userid, forum_groups[g])

            for g in remgroups:
                Phpbb3Manager.__remove_user_from_group(userid, forum_groups[g])

    @staticmethod
    def remove_group(username, group):
        logger.debug(f"Removing phpbb user {username} from group {group}")
        cursor = connections['phpbb3'].cursor()
        userid = Phpbb3Manager.__get_user_id(username)
        if userid is not None:
            groupid = Phpbb3Manager.__get_group_id(group)

            if userid:
                if groupid:
                    try:
                        cursor.execute(Phpbb3Manager.SQL_REMOVE_USER_GROUP, [userid, groupid])
                        logger.info(f"Removed phpbb user {username} from group {group}")
                    except:
                        logger.exception(
                            "Exception prevented removal of phpbb user {} with id {} from group {} with id {}".format(
                                username, userid, group, groupid))
                        pass

    @staticmethod
    def check_user(username):
        logger.debug("Checking phpbb username %s" % username)
        cursor = connections['phpbb3'].cursor()
        cursor.execute(Phpbb3Manager.SQL_USER_ID_FROM_USERNAME, [Phpbb3Manager.__santatize_username(username)])
        row = cursor.fetchone()
        if row:
            logger.debug("Found user %s on phpbb" % username)
            return True
        logger.debug("User %s not found on phpbb" % username)
        return False

    @staticmethod
    def update_user_password(username, characterid, password=None):
        logger.debug("Updating phpbb user %s password" % username)
        cursor = connections['phpbb3'].cursor()
        if not password:
            password = Phpbb3Manager.__generate_random_pass()
        if Phpbb3Manager.check_user(username):
            pwhash = Phpbb3Manager.__gen_hash(password)
            logger.debug(
                f"Proceeding to update phpbb user {username} password with pwhash starting with {pwhash[0:5]}")
            cursor.execute(Phpbb3Manager.SQL_UPDATE_USER_PASSWORD, [pwhash, username])
            Phpbb3Manager.__add_avatar(username, characterid)
            logger.info("Updated phpbb user %s password." % username)
            return password
        logger.error("Unable to update phpbb user %s password - user not found on phpbb." % username)
        return ""

    @staticmethod
    def __update_user_info(username, email, password):
        logger.debug(
            f"Updating phpbb user {username} info: username {email} password of length {len(password)}")
        cursor = connections['phpbb3'].cursor()
        try:
            cursor.execute(Phpbb3Manager.SQL_DIS_USER, [email, password, username])
            logger.info("Updated phpbb user %s info" % username)
        except:
            logger.exception("Unable to update phpbb user %s info." % username)
            pass



from django.db.models.signals import (
    m2m_changed,
    post_save,
    pre_delete,
    pre_save
)
from django.urls import reverse
from unittest import mock

MODULE_PATH = 'allianceauth.authentication'


def patch(target, *args, **kwargs):
    return mock.patch(f'{MODULE_PATH}{target}', *args, **kwargs)


def get_admin_change_view_url(obj: object) -> str:
    """returns URL to admin change view for given object"""
    return reverse(
        'admin:{}_{}_change'.format(
            obj._meta.app_label, type(obj).__name__.lower()
        ),
        args=(obj.pk,)
    )


def get_admin_search_url(ModelClass: type) -> str:
    """returns URL to search URL for model of given object"""
    return '{}{}/'.format(
        reverse('admin:app_list', args=(ModelClass._meta.app_label,)),
        ModelClass.__name__.lower()
    )



from django import forms
from django.contrib.auth.models import Group
from django.core.exceptions import ValidationError
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _

from .models import ReservedGroupName


class GroupAdminForm(forms.ModelForm):
    def clean_name(self):
        my_name = self.cleaned_data['name']
        if ReservedGroupName.objects.filter(name__iexact=my_name).exists():
            raise ValidationError(
                _("This name has been reserved and can not be used for groups."),
                code='reserved_name'
            )
        return my_name


class ReservedGroupNameAdminForm(forms.ModelForm):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields['created_by'].initial = self.current_user.username
        self.fields['created_at'].initial = _("(auto)")

    created_by = forms.CharField(disabled=True)
    created_at = forms.CharField(disabled=True)

    def clean_name(self):
        my_name = self.cleaned_data['name'].lower()
        if Group.objects.filter(name__iexact=my_name).exists():
            raise ValidationError(
                _("There already exists a group with that name."), code='already_exists'
            )
        return my_name

    def clean_created_at(self):
        return now()



import requests
from allianceauth.utils.testing import NoSocketsTestCase, SocketAccessError


class TestNoSocketsTestCase(NoSocketsTestCase):
    def test_raises_exception_on_attempted_network_access(self):

        with self.assertRaises(SocketAccessError):
            requests.get("https://www.google.com")



# Generated by Django 3.2.8 on 2021-10-20 05:22

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('authentication', '0017_remove_fleetup_permission'),
    ]

    operations = [
        migrations.AlterField(
            model_name='state',
            name='name',
            field=models.CharField(max_length=32, unique=True),
        ),
    ]
